// -*- C++ -*-

/*

  The Hoard Multiprocessor Memory Allocator
  www.hoard.org

  Author: Emery Berger, http://www.emeryberger.com
  Copyright (c) 1998-2020 Emery Berger

  See LICENSE.md file at the top-level directory of this distribution
  and at http://github.com/emeryberger/Hoard.

*/


// Use of this source code is governed by the Apache License,
// Version 2.0, that can be found in the LICENSE file.

// Hoard version X.Y.0 = X0Y00

#define HOARD_MAJOR 3
#define HOARD_MINOR 13
#define HOARD_PATCH 0

#define HOARD_VERSION_STRING "3.13.0"
#define HOARD_VERSION_NUMBER (HOARD_MAJOR * 10000 + HOARD_MINOR * 100 + HOARD_PATCH)
